
/*
    pbrt source code is Copyright(c) 1998-2016
                        Matt Pharr, Greg Humphreys, and Wenzel Jakob.

    This file is part of pbrt.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    - Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
    IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

#if defined(_MSC_VER)
#define NOMINMAX
#pragma once
#endif

#ifndef PBRT_LIGHTS_PROJECTION_H
#define PBRT_LIGHTS_PROJECTION_H

// lights/projection.h*
#include "pbrt.h"
#include "light.h"
#include "shape.h"
#include "mipmap.h"

namespace pbrt {

// ProjectionLight Declarations
class ProjectionLight : public Light {
  public:
    // ProjectionLight Public Methods
    #if RENDER_MUTABLE
    ProjectionLight(const Transform &LightToWorld,
                    MediumInterface &medium, const Spectrum &I,
                    const std::string &texname, Float fov);
    #else
    ProjectionLight(const Transform &LightToWorld,
                    const MediumInterface &medium, const Spectrum &I,
                    const std::string &texname, Float fov);
    #endif
    Spectrum Sample_Li(const Interaction &ref,
                       const Point2f &u,
                       Vector3f *wi,
                       Float *pdf,
                       VisibilityTester *vis) const;
    virtual Spectrum Sample_Pos(const Point2f& u,
                                Point3f* p,
                                Float* pdf) const;
    Spectrum Projection(const Vector3f &w) const;
    Spectrum Power() const;
    Float Pdf_Li(const Interaction &, const Vector3f &) const;
    Spectrum Sample_Le(const Point2f &u1, const Point2f &u2, Float time,
                       Ray *ray, Normal3f *nLight, Float *pdfPos,
                       Float *pdfDir) const;
    void Pdf_Le(const Ray &, const Normal3f &, Float *pdfPos,
                Float *pdfDir) const;

  private:
    // ProjectionLight Private Data
    std::unique_ptr<MIPMap<RGBSpectrum>> projectionMap;
    const Point3f pLight;
    const Spectrum I;
    Transform lightProjection;
    Float hither, yon;
    Bounds2f screenBounds;
    Float cosTotalWidth;
};

#if RENDER_MUTABLE
std::shared_ptr<ProjectionLight> CreateProjectionLight(
    const Transform &light2world, Medium *medium,
    const ParamSet &paramSet);
#else
std::shared_ptr<ProjectionLight> CreateProjectionLight(
    const Transform &light2world, const Medium *medium,
    const ParamSet &paramSet);
#endif

}  // namespace pbrt

#endif  // PBRT_LIGHTS_PROJECTION_H
